// Change log
/*
7_4
* made it easy to change pack set up (manny)

7_3
* Multiplication factor for measuredVoltageRaw set to 1.92 (1.897) after calibration measurement in readSerial()

7_2
* Changed CAN car oscillator frequency to 16MHz (8MHz)

7_1
* Simplified and adjusted(measured voltage ref) calculation in readSerial()
* Adjusted Contactor control in loop()

7_0
* Changed pinout to work with BMS_5 HW
* Added voltageSensPowerPIN and set LOW (ON) in setup
* Added readSerial(), to read contactorVoltageRaw from ATTiny()
* Moved Contactor control to loop() (setup()) and added voltage check before main contactor on
* Based on 6_16

*/

// Include libraries
#include "mcp_can.h"
#include "SPI.h"
#include "EEPROM.h"


// Constant declaration
// status MASK
const uint8_t statusACAvailable				= 0x01;
const uint8_t statusLowVoltage		 		= 0x02;
const uint8_t statusVeryLowVoltage 		= 0x04;
const uint8_t statusAllowCharging	 		= 0x08;
const uint8_t statusFalseData			 		= 0x10;
const uint8_t statusDCContactorClosed = 0x20;
//const uint8_t statusLostCom						= 0x40;
const uint8_t statusContactorError		= 0x80;

// errorByte MASK
const uint8_t initError 			= 0x01;
const uint8_t mismatchBefore 	= 0x02;
const uint8_t mismatchAfter 	= 0x04;
const uint8_t receiveBuffer 	= 0x08;

// Battery constants
const uint8_t numberOfModules = 3; // change
const uint8_t numberOfCellsPerModule = 6; // change
const uint8_t numberOfCells = numberOfModules*numberOfCellsPerModule; // change 96
const uint32_t fullCapacity = 3664281600;  // 24.85 * 1024 * 1000 * 3600 / 25 = (24.85 * 147456000)
const int16_t highVoltageCutOff = 4200; // Define high voltage cut off, in mV
const int16_t lowVoltageWarning = 3500; // Define low voltage warning, in mV
const int16_t lowVoltageCutOff = 3200; // Define low voltage cut off, in mV
const int16_t voltageAllowance = 10; // Define the accuracy the balancing algorithm can balance it to, in mV
const int16_t startShuntVoltage = 4050; // Define start shunting voltage, in mV
const uint8_t Message_length = (5+(2*numberOfModules));//Message length (5 + 2 x n = 29)

// Other constants
//const float chargeCompensationFactor = 0.988; // Charge current compensation
const uint8_t balanceCellCounterTurnAround = 40; // Every 10s (40 x 250ms)


// Variable declaration
uint16_t cellVoltage[numberOfCells]; //change 96
int8_t cellBlockTemp[8];
int16_t dieTemperature[numberOfModules]; //change 12
uint32_t totalCapacity;
uint8_t dataReady;
boolean currentSensorsCalibrated = false;
uint16_t lowCurrentZeroCalibration = 512;
uint16_t highCurrentZeroCalibration = 512;
volatile boolean savedToEEPROM = false;


// I/O-PINS
const uint8_t supplyLossPIN = 2; // D2, Int0
const uint8_t ACAvailable = 14; // A0
const uint8_t preChargeRelayPIN = 15; // A1
const uint8_t mainRelayPIN = 16; // A2
const uint8_t SPI_MAX_CS_PIN = 10; // D10, ChipSelect PIN for MAX17841
const uint8_t SPI_MCP_CS_PIN = 8; // D8, CS PIN for MCP2515 CAN module
const uint8_t lowCurrentPIN = 6; // A6 for analogRead()
const uint8_t highCurrentPIN = 7; // A7 for analogRead()
const uint8_t voltageSensPowerPIN = 19; // A5
// SPI-bus
// CLK, SCK - PIN 13
// MISO, SDO - PIN 12
// MOSI, SDI - PIN 11


// SPI setup
SPISettings MAX17841(4000000, MSBFIRST, SPI_MODE0);

// CAN setup
MCP_CAN CAN(SPI_MCP_CS_PIN);  // Set CS pin


/********
* SETUP *
********/
void setup()
{
  pinMode(ACAvailable, INPUT);
  pinMode(supplyLossPIN, INPUT_PULLUP);
  
  pinMode(preChargeRelayPIN, OUTPUT);
  digitalWrite(preChargeRelayPIN, LOW);
  pinMode(mainRelayPIN, OUTPUT);
  digitalWrite(mainRelayPIN, LOW);
  pinMode(voltageSensPowerPIN, OUTPUT);
  digitalWrite(voltageSensPowerPIN, LOW); // Supply ON
  
  pinMode(SPI_MAX_CS_PIN, OUTPUT);
  digitalWrite(SPI_MAX_CS_PIN, HIGH); // No transaction
  
  pinMode(SPI_MCP_CS_PIN, OUTPUT);

	Serial.begin(115200);
	
	initCAN();
	
	for(int i=0; i<10 && !currentSensorsCalibrated; i++){
		delay(100);
		calibrateCurrentSensor();
	}

	dataReady = 0x02; // Reset data ready
		
	Serial.println("Daisy chain init");
  daisyChainInit();
		
  Serial.println("Set all slaves");
  setupSlaves();
  
  Serial.println("Read capacity from EEPROM");
  for(int i=0; i<4; i++)
	{
		uint32_t cap = EEPROM.read(i);
		totalCapacity |= (cap << (i*8));
	}
	
	digitalWrite(preChargeRelayPIN, HIGH);
	/*delay(500); // Precaution, time to 95% is 144ms (-1,2mF x 40ohm x ln(1 - 0.95))
	// Check current sensor for current peak during precharge as extra precaution
	digitalWrite(mainRelayPIN, HIGH);
	delay(100);
	digitalWrite(preChargeRelayPIN, LOW);*/
	
	attachInterrupt(digitalPinToInterrupt(supplyLossPIN), saveCapacityToEEPROM, RISING);
}


/*******
* LOOP *
*******/
void loop()
{
  static uint32_t lastMillisMain = millis();  // Initial start time
  static uint32_t lastMillisPrint = millis();  // Initial start time
  static uint32_t lastMillisContactor = millis();  // Initial start time
	static uint8_t print = 0;
	static uint8_t statusContactor = 0x00;
	uint8_t status = 0x00;
	uint16_t compareVoltage = 0;
	static uint16_t batteryVoltage = 0;
	
	readCAN();
	
	if(digitalRead(preChargeRelayPIN))
	{
		compareVoltage = readSerial();
		
		if((batteryVoltage > 300) && (compareVoltage > batteryVoltage) && (millis() - lastMillisContactor > 500))
		{
			digitalWrite(mainRelayPIN, HIGH);
			lastMillisContactor = millis();
			Serial.print("PreC done, B: ");
			Serial.print(batteryVoltage);
			Serial.print(" C: ");
			Serial.println(compareVoltage);
		}
		
		if(digitalRead(mainRelayPIN) && (millis() - lastMillisContactor > 100))
		{
			digitalWrite(preChargeRelayPIN, LOW);
			digitalWrite(voltageSensPowerPIN, HIGH); // Supply OFF
			statusContactor |= statusDCContactorClosed;
			Serial.println("Contactors ON");
		}
		
		if(millis() - lastMillisContactor > 2000) // Error
		{
			digitalWrite(preChargeRelayPIN, LOW);
			digitalWrite(voltageSensPowerPIN, HIGH); // Supply OFF
			statusContactor |= statusContactorError;
			Serial.println("Contactor Error");
		}
	}
	
	status |= statusContactor;
			
	if (millis() - lastMillisMain >= 250) // Test if 250ms has elapsed
  {
		uint16_t timeSinceLast;
		int16_t instantCurrent;
				
		timeSinceLast = millis() - lastMillisMain;
		lastMillisMain = millis();  // Reset main timer
	
		instantCurrent = readCurrent();
		changeCapacity(instantCurrent, timeSinceLast);
				
		measureCellData();
	
		if(!(digitalRead(ACAvailable))) // Check if AC is available
		{
			status |= statusACAvailable;
		}
		
		if(dataReady == 0x02)
		{
			static boolean measureVoltages = true;
			static uint8_t counter = 0;
			static uint8_t highCell = highestCell();
			static uint8_t lowCell = lowestCell();
			static uint32_t voltage = totalVoltage();
			static uint16_t highestCellVoltage = cellVoltage[highCell];
			static uint16_t lowestCellVoltage = cellVoltage[lowCell];
      static uint8_t staticStatus = 0x00;
			
			if((status & statusACAvailable) && (highestCellVoltage > startShuntVoltage))
			{
				measureVoltages = balanceCells(lowestCellVoltage);
			}
			
			counter++;

      if(counter > (balanceCellCounterTurnAround + 6)) // Error handling 
      {
        counter = 0;
        writeAllSlaves(0x1A, 0x0000, true); // Turn off cell balancing
        measureVoltages = true;
        Serial.println("Balancing error");
      }
      
			if(measureVoltages)
			{
				counter = 0;
				highCell = highestCell();
				lowCell = lowestCell();
				voltage = totalVoltage();
				highestCellVoltage = cellVoltage[highCell];
				lowestCellVoltage = cellVoltage[lowCell];
				
				staticStatus = checkCells(highestCellVoltage, lowestCellVoltage);
				
				batteryVoltage = voltage / 1000; // Total voltage in V
			}

      status |= staticStatus;
      
			sendDataCAN(instantCurrent, voltage, highCell, lowCell, highestCellVoltage, lowestCellVoltage, status);
		}
		else // Error
		{
			status |= statusFalseData;
			sendDataCAN(instantCurrent, 0, 0, 0, 0, 0, status);
			writeAllSlaves(0x1A, 0x0000, true); // Turn off cell balancing
      Serial.println("Data not ready");
		}
		
		print = 1; // Set print to non-zero value
		lastMillisPrint = millis();  // Reset print timer
	}
	
	if((millis() - lastMillisPrint >= 5) && print) // Test if 5ms has lapsed AND print is non-zero
  {
		lastMillisPrint = millis();  // Reset print timer
    print = printData();
	}
	
	if(savedToEEPROM)
	{
		Serial.println("Saved to EEPROM, FORCEPOR");
		savedToEEPROM = false;
	}
}


/*****************************************
* Calibrates current sensors zero offset *
*****************************************/
void calibrateCurrentSensor()
{
	if(currentSensorsCalibrated) {
		return;
	}

	// Calibrate current sensor zero offsets before HV battery is enabled
	Serial.println(F("Current sens calib"));
	
	uint16_t lowCurrent_array[8];
	uint16_t highCurrent_array[8];
	int16_t lowCurrent0 = analogRead(lowCurrentPIN);
	int16_t highCurrent0 = analogRead(highCurrentPIN);
	int16_t lowCurrent1 = 0;
	int16_t highCurrent1 = 0;
	const uint8_t require_stable_count = 8;
	uint8_t stable_count = 0;
	while(stable_count < require_stable_count){
		delay(10);
		lowCurrent1 = analogRead(lowCurrentPIN);
		highCurrent1 = analogRead(highCurrentPIN);
		if(abs(lowCurrent1 - lowCurrent0) >= 10 || abs(highCurrent1 - highCurrent0) >= 10){
			break;
		}
		lowCurrent_array[stable_count] = lowCurrent1;
		highCurrent_array[stable_count] = highCurrent1;
		stable_count++;
	}
	if(lowCurrent1 < 491 || lowCurrent1 > 531 || highCurrent1 < 491 || highCurrent1 > 531){ // 1023 / 2 +- 20 bits
		Serial.print(F("Current sens calib NOK: "));
		Serial.print(lowCurrent1);
		Serial.print(F(", "));
		Serial.print(highCurrent1);
		Serial.print(" (range)");
		Serial.println();
		return;
	}
	if(stable_count < require_stable_count){
		Serial.print(F("Current sens calib NOK: "));
		Serial.print(lowCurrent1);
		Serial.print(F(", "));
		Serial.print(highCurrent1);
		Serial.print(" (unstable)");
		Serial.println();
		return;
	}
	
	uint16_t lowCurrent_sum = 0;
	uint16_t highCurrent_sum = 0;
	for(int i = 0; i < 8; i++) {
		lowCurrent_sum += lowCurrent_array[i];
		highCurrent_sum += highCurrent_array[i];
	}
	
	lowCurrentZeroCalibration = (lowCurrent_sum + 10) / 8;
	highCurrentZeroCalibration = (highCurrent_sum + 10) / 8;
	
	Serial.print(F("Calib offsets: Low, high: "));
	Serial.print(lowCurrentZeroCalibration);
	Serial.print(F(", "));
	Serial.print(highCurrentZeroCalibration);
	Serial.println();

	currentSensorsCalibrated = true;
}


/********************************************************
* Calculates total battery voltage, returns value in mV *
********************************************************/
uint32_t totalVoltage()
{
  uint32_t totV = (uint32_t)0;
  
  for (int i=0; i<numberOfCells; i++)
	{
		totV += cellVoltage[i];  
	}
	
  return totV;
}


/***********************************************************************
* Calculates the highest cell of the battery pack, returns value in mV *
***********************************************************************/
uint8_t highestCell()
{
  uint16_t voltageHighest = 0;
  uint8_t cellNumber;
  
  for (int i=0; i<numberOfCells; i++)
	{
    if (cellVoltage[i] > voltageHighest)
		{
      voltageHighest = cellVoltage[i];
      cellNumber = i;
		}
	}
	
  return cellNumber;
}


/**********************************************************************
* Calculates the lowest cell of the battery pack, returns value in mV *
**********************************************************************/
uint8_t lowestCell()
{
  uint16_t voltageLowest = 5000;
  uint8_t cellNumber;
  
  for(int i=0; i<numberOfCells; i++)
  {
    if (cellVoltage [i] < voltageLowest)
    {
      voltageLowest = cellVoltage[i];
      cellNumber = i;
    }
  }
  
  return cellNumber;
}


/********************************************
* Check status of cells and allows charging *
********************************************/
uint8_t checkCells(uint16_t highestCellVoltage, uint16_t lowestCellVoltage)
{
	uint8_t status = 0x00;
	
	if (highestCellVoltage < highVoltageCutOff) // Check if highest cell is below max allowed voltage
	{
		status |= statusAllowCharging; // Allow charging
	}
	
	if (lowestCellVoltage < lowVoltageWarning) // Check if lowest cell has reached min prefered voltage
	{
		status |= statusLowVoltage; // Warn the driver
		
		if (lowestCellVoltage < lowVoltageCutOff) // Check if lowest cell has reached min allowed voltage
		{
			status |= statusVeryLowVoltage; // Stop car
		}
	}
	
	return status;
}


/************************************************************************
* Reads current sensor, returns value in A x 1024 / 25 to keep accuracy *
************************************************************************/
int16_t readCurrent()
{
	int16_t current;
	int16_t currentLow;
	int16_t currentHigh;
		
	currentLow = analogRead(lowCurrentPIN) - lowCurrentZeroCalibration; 
	currentHigh = analogRead(highCurrentPIN) - highCurrentZeroCalibration; 

  /*Serial.print(currentLow);
  Serial.print("\t");
  Serial.println(currentHigh);*/
	
	if (abs(currentLow) > 350) // Check to use high or low current range sensor //change
	{
		current = 50 * currentHigh; // Current (A x 1024 / 25) //change
	}
	else
	{
		current = 2 * currentLow; // Current (A x 1024 / 25) //change
	}
	
	return current;
}


/***********************************
* Calculates the new totalCapacity *
***********************************/
void changeCapacity(int16_t current, uint16_t lapsedTime)
{
	uint32_t capacityChange;
	uint16_t absCurrent;
		
	if (current < 0) // Discharging
	{
		if (current > -8) // If negative current is less than 0.2A, set absCurrent to 0A, i.e. no capacity change
		{
			absCurrent = 0;
		}
		else
		{
			absCurrent = -(current);
		}
		
		capacityChange = (uint32_t)absCurrent * lapsedTime;
		totalCapacity -= capacityChange; // Stored battery capacity (Ah x 1024 x 1000 x 3600 / 25)	
	}
	else // Charging
	{
		if (current < 8) // If positive current is less than 0.2A, set absCurrent to 0A, i.e. no capacity change
		{
			absCurrent = 0;
		}
		else
		{
			absCurrent = current;
		}
		
		//capacityChange = (uint32_t)absCurrent * lapsedTime * chargeCompensationFactor;
		capacityChange = (uint32_t)absCurrent * lapsedTime;
		totalCapacity += capacityChange; // Stored battery capacity (Ah x 1024 x 1000 x 3600 / 25)
	}
}


/*****************************************************************************************
* Calculates the highest cell temperature of the battery pack, returns value in degree C *
*****************************************************************************************/
int8_t highTempCell()
{
  int8_t highestTemp = -128;
  
  for (int i=0; i<7; i++)
	{
    if (cellBlockTemp[i] > highestTemp)
		{
      highestTemp = cellBlockTemp[i];
		}
	}
	
  return highestTemp;
}


/****************************************************************************************
* Calculates the lowest cell temperature of the battery pack, returns value in degree C *
****************************************************************************************/
int8_t lowTempCell()
{
  int8_t lowestTemp = 127;
  
  for (int i=0; i<7; i++)
	{
    if (cellBlockTemp[i] < lowestTemp)
		{
      lowestTemp = cellBlockTemp[i];
		}
	}
	
  return lowestTemp;
}


/*****************************************************************************************
* Calculates the highest die temperature of the slave modules, returns value in degree C *
*****************************************************************************************/
int8_t highTempDie()
{
  int8_t highestTemp = -128;
  
  for (int i=0; i<7; i++)
	{
    if (dieTemperature[i] > highestTemp)
		{
      highestTemp = dieTemperature[i];
		}
	}
	
  return highestTemp;
}


/********************************
* Sends out battery data on CAN *
********************************/
void sendDataCAN(int16_t current, int32_t voltage, uint8_t highCell, uint8_t lowCell, uint16_t highestCellVoltage, uint16_t lowestCellVoltage, uint8_t status)
{
	int16_t currentLCD = (int32_t)current * 25 / 102.4; // Current in dA
	uint16_t voltageLCD = voltage / 100; // Voltage in dV
	uint16_t capacityLCD = totalCapacity / 14745600; // Capacity in dAh
	
	int8_t cellTempMin = lowTempCell();
	int8_t cellTempMax = highTempCell();
  int8_t dieTempMax = highTempDie();
	
	uint8_t statusToCharger = setCharging(status);	
		
	uint8_t mainBMS[8] = {highByte(currentLCD), lowByte(currentLCD), highByte(voltageLCD), lowByte(voltageLCD), highByte(capacityLCD), lowByte(capacityLCD), status, statusToCharger};
  uint8_t highLowBMS[8] = {(highCell + 1), highByte(highestCellVoltage), lowByte(highestCellVoltage), (lowCell + 1), highByte(lowestCellVoltage), lowByte(lowestCellVoltage), 0x00, 0x00};
  uint8_t tempBMS[8] = {cellTempMin, cellTempMax, cellBlockTemp[7], dieTempMax, statusToCharger, 0x00, 0x00, 0x00};
	
  CAN.sendMsgBuf(0x100, 0, 8, mainBMS);
  CAN.sendMsgBuf(0x101, 0, 8, highLowBMS);
  CAN.sendMsgBuf(0x102, 0, 8, tempBMS);
  
  Serial.print(voltageLCD/10.0, 1);
  Serial.print("\t");
  Serial.print(currentLCD/10.0, 1);
  Serial.print("\t");
  Serial.print(capacityLCD/10.0, 1);
  Serial.print("\t");
  Serial.print(status, HEX);
  Serial.print("\t");
  Serial.println(statusToCharger, HEX);
  
  Serial.print(highCell+1);
  Serial.print(" ");
  Serial.print(highestCellVoltage);
  Serial.print("\t");
  Serial.print(lowCell+1);
  Serial.print(" ");
  Serial.println(lowestCellVoltage);
  
  Serial.print(cellTempMin);
  Serial.print("\t");
  Serial.print(cellTempMax);
  Serial.print("\t");
  Serial.print(cellBlockTemp[7]);
  Serial.print("\t");
  Serial.println(dieTempMax);
}


/******************************************************************
* Checks CAN-bus buffer and if new data is available take action. *
******************************************************************/
void readCAN()
{
	uint32_t rxId;
  uint8_t len;
  uint8_t rxBuf[8];
  uint8_t statusCAN = 0x00;
          
  if(CAN_MSGAVAIL == CAN.checkReceive()) // Check if data is coming
  {
    CAN.readMsgBuf(&rxId, &len, rxBuf); // read data, rxId: message ID, ext: flag extended ID, len: data length, rxBuf: data buf
    
    if(rxId == 0x600) // Status from Charger
    {
      uint8_t chargerState = rxBuf[6];
      chargerState &= 0x0F; // Remove "charger.fail_reason" bits
      
      if(chargerState == 10) // CS_DONE_CHARGING
			{
				totalCapacity = fullCapacity; // Reset battery capacity
			}
    }
  }
}


/********************************************************************
* Reads serial data from ATTiny85 and calculates comparable voltage *
********************************************************************/
uint16_t readSerial()
{
	static uint8_t measuredVoltageRaw = 0;
	boolean newData = false;
	uint16_t compareVoltage = 0;
		
	if(Serial.available() > 0)
  {
		measuredVoltageRaw = Serial.read();
		newData = true;
	}
	
	compareVoltage = 1.92 * measuredVoltageRaw; // 1.92 derived from calibration measurement, 1.897 = 253 * 5.47 / (768 * 0.95) 
	
	if(newData)
	{
		Serial.println(compareVoltage);
		newData = false;
	}
	
	return compareVoltage;
}


/************************************
* Check and set status for charging *
************************************/
uint8_t setCharging(uint8_t status)
{
	uint8_t statusToCharger = 0x00;
	
	if((status & statusAllowCharging) && (status & statusACAvailable))
	{
		statusToCharger |= 0x01;
	}
	else
	{
		statusToCharger &= ~0x01;
	}
	
	if(status & statusDCContactorClosed) 
	{
		statusToCharger |= 0x04;
	}
	else
	{
		statusToCharger &= ~0x04;
	}
	
	if(status & statusACAvailable)
	{
		//sendGearRequestCAN(0x00);
		
		/*uint16_t maxACCurrent = 16; // in A
		uint8_t settingId = 0x00;
		uint16_t oldData = 0x00;
		
		uint8_t chargerSetting[8] = {settingId, highByte(oldData), lowByte(oldData), highByte(maxACCurrent), lowByte(maxACCurrent), 0x00, 0x00, 0x00};
		CAN.sendMsgBuf(0x620, 0, 8, chargerSetting);
		*/
		/*Serial.print(settingId, HEX);
		Serial.print("\t");
		Serial.print(oldData);
		Serial.print("\t");
		Serial.print(maxACCurrent);
		Serial.println(" A");*/
	}
	
	return statusToCharger;
}


/******************************************************
* Measures cell voltages and temperatures from slaves *
******************************************************/
void measureCellData()
{
	uint32_t watchdogTimer = millis();
	
	dataReady = 0x00;
	//writeAllSlaves(0x13, 0x0001, true); // Set SCANCTRL, SCAN
	writeAllSlaves(0x13, 0x0041, true); // Set SCANCTRL, SCAN and 32 oversamples
  
	while(!dataReady) // Wait for data ready
	{		
		readAllSlaves(0x13, true); // Read SCANCTRL to check when data is ready to be read
				
		if((millis() - watchdogTimer) > 15) // Watchdog timer of 15ms
		{
			Serial.println("measureCellData WD timeout");
			return;
		}
	}

  readData(); // Read and store data
  dataReady = 0x02;
}


/*********************
* Balancing of cells *
*********************/
boolean balanceCells(uint16_t lowestCellVoltage)
{
	static uint8_t counter = 1;
	static uint16_t cellToBalanceFilter = 0x0FFF & 0x5555; // Every odd cell
	boolean measureVoltages = false;

  //if(1 == counter) Serial.println("Balance cells");
  
	writeAllSlaves(0x18, 0x1500, true); // Set WATCHDOG timer to 5s

	if(0 == counter)
	{
		cellToBalanceFilter = 0x0FFF & ~cellToBalanceFilter; // Every other cell
		writeAllSlaves(0x1A, 0x0000, true); // Turn off all cell balancing for one cycle to avoid any switch overlap
	}
	
	if(1 == counter)
	{
		measureVoltages = true; // OK to measure cell voltages
	}
	
	for (int i=0; i < numberOfModules; i++) // Go through modules to set cells for balancing
	{
		if(2 == counter) // Only set cells once every counter cycle
		{
			uint16_t cellToBalance = 0x0000;
    
			for (int j=0; j < numberOfCellsPerModule; j++) // Go through cells to set cells for balancing
			{
				uint8_t cellNumber = j+i*numberOfCellsPerModule;
			
				if ((cellVoltage[cellNumber] > startShuntVoltage) && ((int16_t)(cellVoltage[cellNumber] - lowestCellVoltage) > voltageAllowance)) // Ok to balance
				{
//					if (j < 4)
//					{
						cellToBalance |= (0x01 << j);
//					}
//					else if (j >= 4 && j <8)
//					{
//						cellToBalance |= (0x01 << (j+1));
//					}
//					else //Error
//					{
//						cellToBalance = 0x0000;
//					}
				}
			}
		
			cellToBalance &= cellToBalanceFilter; // To avoid enabling adjacent balance switches
			writeAddressedSlave(0x1A, cellToBalance, i, true); // Send request to set cell balance switches
		
			Serial.print(i+1);
			Serial.print(": ");
			//Serial.println(cellToBalance, BIN);
			
			for(int i=0; i < numberOfCellsPerModule; i++)
			{
//				if(4 == i) Serial.print(" ");
//				
//				if(i < 4)
//				{
					Serial.print((uint8_t)((cellToBalance >> i) & 0x01));
//				}
//				else
//				{
//					Serial.print((uint8_t)((cellToBalance >> (i + 1)) & 0x01));
//				}
			}
			Serial.println();
			
		}
		
		if(dieTemperature[i] > 100) // Check die overtemp
		{
			writeAddressedSlave(0x1A, 0x0000, i, true); // Turn off cell balancing for the module with die overtemp
			Serial.print("Die overtemp, module ");
      Serial.println(i+1);
		}
	}
	
	counter++;
	
	if(counter > balanceCellCounterTurnAround)
	{
		counter = 0;
	}
	
	return measureVoltages;
}


/************************************
* Reads data from all slave devices *
************************************/
void readAllSlaves(uint8_t dataRegister, boolean setupDone) // Read all slaves
{
	uint8_t command = 0x03; // READALL
	uint8_t byteList[3] = {command, dataRegister, 0x00};
	uint8_t PEC = calculatePEC(byteList, 3);
	uint8_t readRegisterData[Message_length]; //change
	uint8_t errorByte = 0x00;
	static uint8_t resendCounter = 0;
	//static uint8_t failCounter = 0;
	
	/*if(failCounter >= 4)
	{
		//clearBuffers();
		failCounter = 0; // Reset counter
		Serial.println("Buffers cleared");
	}*/
 
	SPI.beginTransaction(MAX17841);
	
	// Load the READALL command sequence into the load queue
	digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0xC0); // WR_LD_Q SPI command byte (write the load queue)
  SPI.transfer(Message_length); // Message length (5 + 2 x n = 29) //change
  SPI.transfer(command); // READALL command byte
  SPI.transfer(dataRegister); // Register address
  SPI.transfer(0x00); // Data-check byte (seed value = 0x00)
  SPI.transfer(PEC); // PEC byte
  SPI.transfer(0x00); // Alive-counter byte (seed value = 0x00)
  digitalWrite(SPI_MAX_CS_PIN, HIGH); 
  
  transmitQueue();
	
	// Read the receive buffer
	digitalWrite(SPI_MAX_CS_PIN, LOW);
	SPI.transfer(0x93); // RD_NXT_MSG SPI command byte
	
	for(int i=0; i<Message_length; i++) //change
  {
		readRegisterData[i] = SPI.transfer(0x93);
	}
	
	errorByte |= receiveBufferError();
	
	SPI.endTransaction();
	
	// Verify that the device register data is received correctly during the READALL sequence
	if(!((readRegisterData[0] == command) && (readRegisterData[1] == dataRegister)))
  {
		errorByte |= mismatchBefore; 
	}
	if(setupDone)
	{
		uint8_t checkPEC = calculatePEC(readRegisterData, (Message_length-2) ); //change
		// Check check-byte, PEC and alive-counter	
		if(!((readRegisterData[(Message_length-3)] == 0x00) && (readRegisterData[(Message_length-2)] == checkPEC) && (readRegisterData[(Message_length-1)] == numberOfModules))) //change
		{
			errorByte |= mismatchAfter;
			if(readRegisterData[(Message_length-3)] != 0x00) // change
			{
				readAllSlaves(0x02, false); // Read STATUS of all slaves with checks ignored
				writeAllSlaves(0x02, 0x0000, true); // Clear STATUS register
				Serial.println("STATUS cleared");
			}
		}
	}
	
	/*// Print data received
	for(int i=0; i<Message_length; i++)//change
  {
		Serial.print(readRegisterData[i], HEX);
		Serial.print(" ");
	}
	Serial.println();*/

	if(errorByte) // Error
	{
    resendCounter++;
    Serial.println(errorByte, HEX);
    errorByte &= 0x00; // Clear errors
    Serial.println("errorByte cleared readAll");
    
    if(resendCounter > 2)
    {
			Serial.println("READALL fail");
			resendCounter = 0;
			//failCounter++;
			return;
		}
    
    delay(1);
    readAllSlaves(dataRegister, setupDone); // Resend READALL
  }
  else // No errors, clear counters and store data received
  {
		resendCounter = 0; // Reset counter
		//failCounter = 0; // Reset counter
		
		if((dataRegister >= 0x20) && (dataRegister <= 0x2B)) // Cell voltage measurements
		{
			storeCellVoltage(dataRegister, readRegisterData);
		}
  
		if(dataRegister == 0x2D) // Aux voltage measurements (external temperature sensors)
		{
			storeCellTemperature(readRegisterData);
		}
		
		if(dataRegister == 0x50) // Die temperature measurements
		{
			storeDieTemperature(readRegisterData);
		}
		
		if(dataRegister == 0x13) // Read SCANCTRL to check if data is ready to be read
		{
			checkDataReady(readRegisterData);
		}
		
		if(dataRegister == 0x02) // Read STATUS and print content
		{
			// Print data received
			for(int i=0; i<Message_length; i++)
			{
				Serial.print(readRegisterData[i], HEX);
				Serial.print(" ");
			}
			Serial.println();
		}
	}  
}


/*****************************************
* Reads data from addressed slave device *
*****************************************/
void readAddressedSlave(uint8_t dataRegister, uint8_t address, boolean setupDone) // Read addressed slave
{
	uint8_t command = 0x05; // READDEVICE
	command |= (address << 3);
	uint8_t byteList[3] = {command, dataRegister, 0x00};
	uint8_t PEC = calculatePEC(byteList, 3);
	uint8_t readRegisterData[7];
	uint8_t errorByte = 0x00;
	static uint8_t resendCounter = 0;
	//static uint8_t failCounter = 0;
	
	/*if(failCounter >= 4)
	{
		//clearBuffers();
		failCounter = 0; // Reset counter
		Serial.println("Buffers cleared");
	}*/
		
	SPI.beginTransaction(MAX17841);
	
	// 1, Load the READDEVICE command sequence into the load queue
	digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0xC0); // WR_LD_Q SPI command byte (write the load queue)
  SPI.transfer(0x07); // Message length (5 + 2 x n = 29)
  SPI.transfer(command); // READALL command byte
  SPI.transfer(dataRegister); // Register address
  SPI.transfer(0x00); // Data-check byte (seed value = 0x00)
  SPI.transfer(PEC); // PEC byte
  SPI.transfer(0x00); // Alive-counter byte (seed value = 0x00)
  digitalWrite(SPI_MAX_CS_PIN, HIGH); 
  
  transmitQueue();
	
	// 4, Read the receive buffer
	digitalWrite(SPI_MAX_CS_PIN, LOW);
	SPI.transfer(0x93); // RD_NXT_MSG SPI command byte
	
	for(int i=0; i<7; i++)
  {
		readRegisterData[i] = SPI.transfer(0x93);
	}
	
	errorByte |= receiveBufferError();
	
	SPI.endTransaction();
	
	// Verify that the device register data is received correctly during the READDEVICE sequence
	if(!((readRegisterData[0] == command) && (readRegisterData[1] == dataRegister)))
  {
		errorByte |= mismatchBefore; 
	}
	if(setupDone)
	{
		uint8_t checkPEC = calculatePEC(readRegisterData, 5);
		// Check check-byte, PEC and alive-counter		
		if(!((readRegisterData[4] == 0x00) && (readRegisterData[5] == checkPEC) && (readRegisterData[6] == 0x01)))
		{
			errorByte |= mismatchAfter;
			if(readRegisterData[4] != 0x00)
			{
				readAddressedSlave(0x02, address, false); // Read STATUS of addressed slave with checks ignored
				writeAddressedSlave(0x02, 0x0000, address, true); // Clear STATUS register
				Serial.println("STATUS cleared");
			}
		}
	}
	
	/*// Print data received
	for(int i=0; i<7; i++)
  {
		Serial.print(readRegisterData[i], HEX);
		Serial.print(" ");
	}
	Serial.println();*/
	
	if(errorByte) // Error
	{
    resendCounter++;
    Serial.println(errorByte, HEX);
    errorByte &= 0x00; // Clear errors
    Serial.println("errorByte cleared readAddr");
    
    if(resendCounter > 2)
    {
			Serial.println("READEVICE fail");
			resendCounter = 0;
			//failCounter++;
			return;
		}
    
    delay(1);
    readAddressedSlave(dataRegister, address, setupDone); // Resend READDEVICE
  }
  else // No errors, clear counters and store data received
  {
		resendCounter = 0; // Reset counter
		//failCounter = 0; // Reset counter
		
		if(dataRegister == 0x02) // Read STATUS and print content
		{
			// Print data received
			for(int i=0; i<7; i++)
			{
				Serial.print(readRegisterData[i], HEX);
				Serial.print(" ");
			}
			Serial.println();
		}
	} 
}


/***********************************
* Writes data to all slave devices *
***********************************/
void writeAllSlaves(uint8_t dataRegister, uint16_t data, boolean setupDone)
{
	uint8_t command = 0x02; // WRITEALL
	uint8_t byteList[4] = {command, dataRegister, lowByte(data), highByte(data)};
	uint8_t PEC = calculatePEC(byteList, 4);
	uint8_t readRegisterData[6];
	uint8_t errorByte = 0x00;
	static uint8_t resendCounter = 0;
	//static uint8_t failCounter = 0;
	
	/*if(failCounter >= 4)
	{
		//clearBuffers();
		failCounter = 0; // Reset counter
		Serial.println("Buffers cleared");
	}*/
	
	SPI.beginTransaction(MAX17841);
	
	// 1, Load the WRITEALL command sequence into the load queue
	digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0xC0); // WR_LD_Q SPI command byte (write the load queue)
  SPI.transfer(0x06); // Message length
  SPI.transfer(command); // WRITEALL command byte
  SPI.transfer(dataRegister); // Register address
  SPI.transfer(lowByte(data)); // LS byte of register data to be written
  SPI.transfer(highByte(data)); // MS byte of register data to be written
  SPI.transfer(PEC); // PEC byte
  SPI.transfer(0x00); // Alive-counter byte (seed value = 0x00)
  digitalWrite(SPI_MAX_CS_PIN, HIGH); 
  
  transmitQueue();
	
	// 4, Read the receive buffer
	digitalWrite(SPI_MAX_CS_PIN, LOW);
	SPI.transfer(0x93); // RD_NXT_MSG SPI command byte
	
	for(int i=0; i<6; i++)
  {
		readRegisterData[i] = SPI.transfer(0x93);
	}
  digitalWrite(SPI_MAX_CS_PIN, HIGH);
	
	errorByte |= receiveBufferError();
  
  SPI.endTransaction();
  
  // Verify that the device register data is what was written during the WRITEALL sequence
  if(!((readRegisterData[0] == command) && (readRegisterData[1] == dataRegister) && (readRegisterData[2] == lowByte(data)) && (readRegisterData[3] == highByte(data)) && (readRegisterData[4] == PEC)))
  {
		errorByte |= mismatchBefore; 
	}
	if(setupDone)
	{
		// Check alive-counter
		if(!(readRegisterData[5] == numberOfModules)) // change
		{
			errorByte |= mismatchAfter;
		}
	}

	/*// Print data received
  for(int i=0; i<6; i++)
  {
    Serial.print(readRegisterData[i], HEX);
    Serial.print(" ");
  }
  Serial.println();*/
  
  if(errorByte) // Error
	{
    resendCounter++;
    Serial.println(errorByte, HEX);
    errorByte &= 0x00; // Clear errors
    Serial.println("errorByte cleared writeAll");
    
    if(resendCounter > 2)
    {
			Serial.println("WRITEALL fail");
			resendCounter = 0;
			//failCounter++;
			return;
		}
    
    delay(1);
    writeAllSlaves(dataRegister, data, setupDone); // Resend WRITEALL
  }
  else // No errors, clear counters
  {
		resendCounter = 0; // Reset counter
		//failCounter = 0; // Reset counter
	}
}


/****************************************
* Writes data to addressed slave device *
****************************************/
void writeAddressedSlave(uint8_t dataRegister, uint16_t data, uint8_t address, boolean setupDone) // Write addressed slave
{
	uint8_t command = 0x04; // WRITEDEVICE
	command |= (address << 3);
	uint8_t byteList[4] = {command, dataRegister, lowByte(data), highByte(data)};
	uint8_t PEC = calculatePEC(byteList, 4);
	uint8_t readRegisterData[6];
	uint8_t errorByte = 0x00;
	static uint8_t resendCounter = 0;
	//static uint8_t failCounter = 0;
	
	/*if(failCounter >= 4)
	{
		//clearBuffers();
		failCounter = 0; // Reset counter
		Serial.println("Buffers cleared");
	}*/
		
	SPI.beginTransaction(MAX17841);
	
	// 1, Load the WRITEDEVICE command sequence into the load queue
	digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0xC0); // WR_LD_Q SPI command byte (write the load queue)
  SPI.transfer(0x06); // Message length
  SPI.transfer(command); // WRITEDEVICE command byte
  SPI.transfer(dataRegister); // Register address
  SPI.transfer(lowByte(data)); // LS byte of register data to be written
  SPI.transfer(highByte(data)); // MS byte of register data to be written
  SPI.transfer(PEC); // PEC byte
  SPI.transfer(0x00); // Alive-counter byte (seed value = 0x00)
  digitalWrite(SPI_MAX_CS_PIN, HIGH); 
  
  transmitQueue();
	
	// 4, Read the receive buffer
	digitalWrite(SPI_MAX_CS_PIN, LOW);
	SPI.transfer(0x93); // RD_NXT_MSG SPI command byte
	
	for(int i=0; i<6; i++)
  {
		readRegisterData[i] = SPI.transfer(0x93);
	}
  digitalWrite(SPI_MAX_CS_PIN, HIGH);
	
	errorByte |= receiveBufferError();
  
  SPI.endTransaction();
  
	// Verify that the device register data is what was written during the WRITEDEVICE sequence
	if(!((readRegisterData[0] == command) && (readRegisterData[1] == dataRegister) && (readRegisterData[2] == lowByte(data)) && (readRegisterData[3] == highByte(data)) && (readRegisterData[4] == PEC)))
  {
		errorByte |= mismatchBefore; 
	}
	if(setupDone)
	{
		// Check alive-counter
		if(!(readRegisterData[5] == 0x01))
		{
			errorByte |= mismatchAfter;
		}
	}

	/*// Print data received
  for(int i=0; i<6; i++)
  {
    Serial.print(readRegisterData[i], HEX);
    Serial.print(" ");
  }
  Serial.println();*/
  
  if(errorByte) // Error
	{
    resendCounter++;
    Serial.println(errorByte, HEX);
    errorByte &= 0x00; // Clear errors
    Serial.println("errorByte cleared writeAddr");
    
    if(resendCounter > 2)
    {
			Serial.println("WRITEDEVICE fail");
			resendCounter = 0;
			//failCounter++;
			return;
		}
    
    delay(1);
    writeAddressedSlave(dataRegister, data, address, setupDone); // Resend WRITEDEVICE
  }
  else // No errors, clear counters
  {
		resendCounter = 0; // Reset counter
		//failCounter = 0; // Reset counter
	}
}


/**********************************************
* Reads data from slaves and stores in arrays *
**********************************************/
void readData() 
{	
	for(int i=0; i<numberOfCellsPerModule; i++) // change
  {
    readAllSlaves((0x20 + i), true); // Read CELL 1-4 of all slaves
  }
//  for(int i=0; i<4; i++) // change
//  {
//    readAllSlaves((0x25 + i), true); // Read CELL 5-8 of all slaves
//  }
  readAllSlaves(0x2D, true); // Read AIN1 of all slaves (Cell temperature)
  readAllSlaves(0x50, true); // Read DIAG (Die temperature) of all slaves
}


/********************
* Prints slave data *
********************/
uint8_t printData() 
{	
	static uint8_t counter = 0;
	counter++;
  
	if(counter < (numberOfModules+1)) //change
	{
		Serial.print(counter);
		Serial.print(": ");
	
		for(int i=0; i<numberOfCellsPerModule; i++)
		{
			Serial.print(cellVoltage[i+(counter-1)*numberOfCellsPerModule]);
			Serial.print(" ");
		}
		Serial.print(dieTemperature[(counter-1)]);
		Serial.println();
	}
  
  if(counter >= (numberOfModules+1)) // change
  {
		for(int i=0; i<numberOfModules; i++)
		{
			Serial.print(cellBlockTemp[i]);
			Serial.print(" ");
		}
		Serial.println();
		
		counter = 0;
	}
	
	return counter;
}


/*********************************************************
* Stores measured cell voltage data in cellVoltage array *
*********************************************************/
void storeCellVoltage(uint8_t dataRegister, uint8_t readRegisterData[29]) 
{	  
	if((dataRegister >= 0x20) && (dataRegister <= 0x28)) // Cell voltage registers 1-4
  {
    uint8_t cellNumberOffset = dataRegister - 0x20;
    
    for(int i=0; i<numberOfModules; i++) //change
		{
			uint16_t measVoltage = ((readRegisterData[(Message_length-4)-i*2] << 8) + readRegisterData[(Message_length-5)-i*2]);
			measVoltage = (measVoltage >> 2) * (uint32_t)5000 / 0x3FFF;
			cellVoltage[i*numberOfCellsPerModule + cellNumberOffset] = measVoltage;
		}
	}
		
//	if((dataRegister >= 0x25) && (dataRegister <= 0x28)) // Cell voltage registers 5-8
//  {
//    uint8_t cellNumberOffset = dataRegister - 0x20 - 1;
//    
//    for(int i=0; i<numberOfModules; i++) //change
//		{
//			uint16_t measVoltage = ((readRegisterData[(Message_length-4)-i*2] << 8) + readRegisterData[(Message_length-5)-i*2]);
//			measVoltage = (measVoltage >> 2) * (uint32_t)5000 / 0x3FFF;
//			cellVoltage[i*8 + cellNumberOffset] = measVoltage;
//		}
//	}
}


/***************************************************************************
* Stores measured temperature sensor voltage data in cellTemperature array *
***************************************************************************/
void storeCellTemperature(uint8_t readRegisterData[29]) 
{	
	int8_t cellTemperature[12];
	
	for(int i=0; i<12; i++)
	{
		if((i == 0) || (i == 7) || (i == 8) || (i == 11)) // No temp sensor connected
    {
      cellTemperature[i] = 0;
    }
    else
    {		
		  uint16_t beta = 3800;
		  uint16_t measTemperature = ((readRegisterData[25-i*2] << 8) + readRegisterData[24-i*2]);
		  measTemperature = (measTemperature >> 4);
      int8_t temperature = beta / (log((float)measTemperature / (4095 - measTemperature)) + beta / 298.15) - 273;
		  cellTemperature[i] = temperature;
    }
	}
	
	cellBlockTemp[0] = cellTemperature[1];
	cellBlockTemp[1] = cellTemperature[2];
	cellBlockTemp[2] = cellTemperature[3];
	cellBlockTemp[3] = cellTemperature[5];
	cellBlockTemp[4] = cellTemperature[6];
	cellBlockTemp[5] = cellTemperature[9];
	cellBlockTemp[6] = cellTemperature[10];
	cellBlockTemp[7] = cellTemperature[4]; // Battery box internal temperature
}


/***************************************************************
* Stores measured die temperature data in dieTemperature array *
***************************************************************/
void storeDieTemperature(uint8_t readRegisterData[Message_length]) 
{	
	for(int i=0; i<numberOfModules; i++)
	{
		uint16_t measTemperature = ((readRegisterData[(Message_length-4)-i*2] << 8) + readRegisterData[(Message_length-5)-i*2]);
		measTemperature = (measTemperature >> 2) * (uint32_t)230700 / 5029581;
		dieTemperature[i] = measTemperature - (int16_t)273;
	}
}


/****************************************************
* Check if data is ready to be read from all slaves *
****************************************************/
void checkDataReady(uint8_t readRegisterData[Message_length]) 
{	
	dataReady = 0x01;
	for(int i=0; i<numberOfModules; i++)
	{
		if(!(readRegisterData[(Message_length-4)-i*2] & 0x80))
		dataReady = 0x00;
	}    
}


/***********************************************************
* PEC Calculation, CRC-8, from MAX17823 datasheet, page 96 *
***********************************************************/
uint8_t calculatePEC(uint8_t byteList[Message_length], uint8_t numberOfBytes)
{
	uint8_t CRCByte = 0;
	uint8_t POLY = 0xB2;
	
	for (int byteCounter = 0; byteCounter < numberOfBytes; byteCounter++)
	{
		CRCByte ^= byteList[byteCounter];
		
		for (int bitCounter = 0; bitCounter < 8; bitCounter++)
		{
			CRCByte = (CRCByte & 0x01) ? ((CRCByte >> 1) ^ POLY) : (CRCByte >> 1);
		}
	}
	
	return CRCByte;
}


/********************
* Slave board setup *
********************/
void setupSlaves()
{
	// Setup
	readAllSlaves(0x01, false); // Read ADDRESS of all slaves
  readAllSlaves(0x02, false); // Read STATUS of all slaves
  writeAllSlaves(0x02, 0x0000, false); // Clear STATUS register
  writeAllSlaves(0x10, 0x0040, false); // Set DEVCFG1, ALIVECNTEN
  //setupDone = true;

  // Enable measurement
  writeAllSlaves(0x12, 0x11FF, true); // Set MEASUREEN, Enable cell voltages 1-4, 6-9 and AUX1 // change
  writeAllSlaves(0x1E, 0x000C, true); // Set TOPCELL, Top cell for measurement is 9 // change
  writeAllSlaves(0x51, 0x0006, true); // Set DIAGCFG, Enable Die temperature measurement
  //writeAllSlaves(0x19, 0x003F, true); // Set ACQCFG, 64x6us settling time before AUX1 measurement
  //writeAllSlaves(0x18, 0x1500, true); // Set WATCHDOG, Set watchdg for cell balancing to 5s
  //writeAllSlaves(0x13, 0x0001, true); // Set SCANCTRL, SCAN
}


/*********************************************************************
* Wait for null messages and then clear transmit and receive buffers *
*********************************************************************/
void clearBuffers()
{
	// Wait for null message to be received
  uint32_t watchdogTimer = millis();
  uint8_t check = 0x01;
  while(check & 0x01) // If RX_Status[0] (RX_Empty_Status) is false, continue. If true, then repeat transaction until false.
  {
		digitalWrite(SPI_MAX_CS_PIN, LOW);
		SPI.transfer(0x01); // Read RX_Status register
		check = SPI.transfer(0x01); // Read RX_Status register
		digitalWrite(SPI_MAX_CS_PIN, HIGH);
		
		if((millis() - watchdogTimer) > 10) // Watchdog timer of 10ms
		{
			Serial.println("clearBuffers WD timeout");
			return;
		}
 	}
	
	// Clear transmit buffer
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0x20); // Clear transmit buffer
  digitalWrite(SPI_MAX_CS_PIN, HIGH);
  
  // Clear receive buffer
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0xE0); // Clear receive buffer
  digitalWrite(SPI_MAX_CS_PIN, HIGH);
}


/********************************************************
* Start transmitting the queue and check receive buffer *
********************************************************/
void transmitQueue()
{
	uint32_t watchdogTimer = millis();
	uint8_t check = 0;
	
	// Start transmitting the loaded sequence from the transmit queue
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0xB0); // WR_NXT_LD_Q SPI command byte (write the next load queue)
  digitalWrite(SPI_MAX_CS_PIN, HIGH);
	
	// Check if a message has been received into the receive buffer
  while(!(check &= 0x12)) // If RX_Status[1] is true, continue. If false, then repeat transaction until true
  {
		// Poll RX_Stop_Status bit
		digitalWrite(SPI_MAX_CS_PIN, LOW);
		SPI.transfer(0x01); // Read RX_Status register
		check = SPI.transfer(0x01); // Read RX_Status register
		digitalWrite(SPI_MAX_CS_PIN, HIGH);
		
		if((millis() - watchdogTimer) > 10) // Watchdog timer of 10ms
		{
			Serial.println("transmitQueue WD timeout");
			return;
		}
	}
  delay(1); // Needed to work, unknown why?
}


/*********************************
* Check for receive buffer error *
*********************************/
uint8_t receiveBufferError()
{
	uint8_t check = 0x00;
	uint8_t errorByte = 0x00;
	
	// Check for receive buffer errors
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0x09); // Read RX_Interrupt_Flags register
  check = SPI.transfer(0x09);
	digitalWrite(SPI_MAX_CS_PIN, HIGH);
   
  if(!(check == 0x00)) // Error
  {
		errorByte |= receiveBuffer; // Set status byte
		
		Serial.print("Error ");
		Serial.println(check, HEX);

		// Clear INT flag register
    digitalWrite(SPI_MAX_CS_PIN, LOW);
    SPI.transfer(0x08); // Write RX_Interrupt_Flags register
    SPI.transfer(0x00); // Clear flags
    digitalWrite(SPI_MAX_CS_PIN, HIGH);
	}
	
	return errorByte;
}


/*******************************************
* UART Daisy-Chain Initialization Sequence *
*******************************************/
void daisyChainInit()
{
	uint8_t check;
	uint8_t data[4];
	uint8_t errorByte = 0x00;
	
	SPI.beginTransaction(MAX17841);

	// 1, Enable Keep-Alive mode
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0x10); // Write Configuration 3 register
  SPI.transfer(0x05); // Set keep-alive period to 160μs
  digitalWrite(SPI_MAX_CS_PIN, HIGH);

  // 2, Enable Rx Interrupt flags for RX_Error and RX_Overflow
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0x04); // Write RX_Interrupt_Enable register
  SPI.transfer(0x88); // Set the RX_Error_INT_Enable and RX_Overflow_INT_Enable bits
  digitalWrite(SPI_MAX_CS_PIN, HIGH);
  
  // 3, Clear receive buffer
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0xE0); // Clear receive buffer
  digitalWrite(SPI_MAX_CS_PIN, HIGH);
  
  // 4, Wake-up UART slave devices
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0x0E); // Write Configuration 2 register
  SPI.transfer(0x30); // Enable Transmit Preambles mode
  digitalWrite(SPI_MAX_CS_PIN, HIGH);
  
  // 5, Wait for all UART slave devices to wake up
  uint32_t watchdogTimer = millis();
  check = 0;
  while(check != 0x21) // If RX_Status = 21h, continue. Otherwise, repeat transaction until true or timeout
  {
    digitalWrite(SPI_MAX_CS_PIN, LOW);
		SPI.transfer(0x01); // Read RX_Status register
		check = SPI.transfer(0x01); // Read RX_Status register
		digitalWrite(SPI_MAX_CS_PIN, HIGH);
		
		if((millis() - watchdogTimer) > 100) // Watchdog timer of 100ms
		{
			Serial.println("UARTSlaveDevicesWakeUp WD timeout");
			return;
		}
	}
	
	// 6, End of UART slave device wake-up period
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0x0E); // Write Configuration 2 register
  SPI.transfer(0x10); // Disable Transmit Preambles mode
  digitalWrite(SPI_MAX_CS_PIN, HIGH);
  
  // 7, Wait for null message to be received
	// 8, Clear transmit buffer
  // 9, Clear receive buffer
  clearBuffers();
  
  // 10, Load the HELLOALL command sequence into the load queue
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0xC0); // WR_LD_Q SPI command byte (write the load queue)
  SPI.transfer(0x03); // Message length
  SPI.transfer(0x57); // HELLOALL command byte
  SPI.transfer(0x00); // Register address (0x00)
  SPI.transfer(0x00); // Initialization address of HELLOALL
  digitalWrite(SPI_MAX_CS_PIN, HIGH);
  
  // 11, Verify contents of the load queue
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0xC1); // RD_LD_Q SPI command byte
  
  for(int i=0; i<4; i++)
  {
		data[i] = SPI.transfer(0xC1);
	}
  
  if(!((data[0] == 0x03) && (data[1] == 0x57) && (data[2] == 0x00) && (data[3] == 0x00)))
  {
		errorByte |= initError;
	}
  digitalWrite(SPI_MAX_CS_PIN, HIGH);
  
  // 12, Transmit HELLOALL sequence
  // 13, Poll RX_Stop_Status bit
  transmitQueue();
	
	// 14, Read the HELLOALL message that propagated through the daisy-chain and was returned back to the ASCI
  digitalWrite(SPI_MAX_CS_PIN, LOW);
  SPI.transfer(0x93); // RD_NXT_MSG SPI transaction
  
  for(int i=0; i<3; i++)
  {
		data[i] = SPI.transfer(0x93);
	}
  
  if(!((data[0] == 0x57) && (data[1] == 0x00) && (data[2] == numberOfModules )))
  {
		errorByte |= initError;
	}
  digitalWrite(SPI_MAX_CS_PIN, HIGH);
  
  // 15, Check for receive buffer errors
  errorByte |= receiveBufferError();
  
  SPI.endTransaction();
  
  if(errorByte) // Error
	{
    Serial.println(errorByte, HEX);
    errorByte &= 0x00; // Clear errors
    Serial.println("errorByte cleared daisyChain");
    daisyChainInit(); // Redo Daisy chain init
  }
}


void initCAN()
{
	uint32_t watchdogTimer = millis();	
	
	while (CAN_OK != CAN.begin(MCP_STDEXT, CAN_250KBPS, MCP_16MHZ))    // init can bus : baudrate = 250k
  { 
		delay(100);
		
		if((millis() - watchdogTimer) > 2000) // Watchdog timer of 2s
		{
			Serial.println("InitCAN WD timeout");
			return;
		}
	}
	
	CAN.setMode(MCP_NORMAL);  // Set operation mode to normal so the MCP2515 sends acks to received data.
  
  CAN.init_Mask(0,0,0x07FF0000);                // Init first mask
  CAN.init_Filt(0,0,0x01210000);                // Init first filter
  CAN.init_Filt(1,0,0x01320000);                // Init second filter
  
  CAN.init_Mask(1,0,0x07FF0000);                // Init second mask 
  CAN.init_Filt(2,0,0x01210000);                // Init third filter
  CAN.init_Filt(3,0,0x01320000);                // Init fouth filter
  CAN.init_Filt(4,0,0x06000000);                // Init fifth filter
  CAN.init_Filt(5,0,0x06000000);                // Init sixth filter
}



/******************************************************************
* Called by interrupt FALLING, to save battery capacity to EEPROM *
******************************************************************/
void saveCapacityToEEPROM()
{
	for(int i=0; i<4; i++)
	{
		uint8_t cap = (totalCapacity >> (i*8)) & 0xFF;
		EEPROM.update(i, cap);
	}
	//Serial.println("Saved");

  writeAllSlaves(0x10, 0x0080, true); // Set DEVCFG1, FORCEPOR
  //Serial.println("FORCEPOR");
  
  savedToEEPROM = true;
}
